/*
 * tkImgGIF.c --
 *
 * A photo image file handler for GIF files. Reads 87a and 89a GIF files.
 * At present there is no write function.
 *
 * Derived from the giftoppm code found in the pbmplus package 
 * and tkImgFmtPPM.c in the tk4.0b2 distribution by -
 *
 * Reed Wade (wade@cs.utk.edu), University of Tennessee
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * This file also contains code from the giftoppm program, which is
 * copyrighted as follows:
 *
 * +-------------------------------------------------------------------+
 * | Copyright 1990, David Koblas.                                     |
 * |   Permission to use, copy, modify, and distribute this software   |
 * |   and its documentation for any purpose and without fee is hereby |
 * |   granted, provided that the above copyright notice appear in all |
 * |   copies and that both that copyright notice and this permission  |
 * |   notice appear in supporting documentation.  This software is    |
 * |   provided "as is" without express or implied warranty.           |
 * +-------------------------------------------------------------------+
 *
 * SCCS: @(#) tkImgGIF.c 1.9 96/07/19 09:57:18
 */

#include "config.h"
#include <tk.h>
#include <stdlib.h>
#include <string.h>
#include "tclcl.h"


#define _debug_                 stderr
#define INTERLACE		0x40
#define LOCALCOLORMAP		0x80
#define BitSet(byte, bit)	(((byte) & (bit)) == (bit))
#define MAXCOLORMAPSIZE		256
#define CM_RED			0
#define CM_GREEN		1
#define CM_BLUE			2
#define MAX_LWZ_BITS		12
#define LM_to_uint(a,b)         (((b)<<8)|(a))
#define ReadOK(buffer,len)	(ReadData(buffer, len, 1) != 0)



class GIFObject : public TclObject {
public:
  /*
   * The format record for the GIF file format:
   */
  
  static int Match _ANSI_ARGS_((char *objectName, char *formatString, 
				int *widthPtr, int *heightPtr));
  static int Read  _ANSI_ARGS_((Tcl_Interp *interp,
				char *objectName, char *formatString,
				Tk_PhotoHandle imageHandle, 
				int destX, int destY,
				int width, int height, 
				int srcX, int srcY));

  GIFObject(const char *string);
  ~GIFObject();

private:
  char *buffer;
  int  size;
  char *nextByte;

  /*
   * Prototypes for local procedures defined in this class:
   */

  void Reset() { nextByte = buffer; }
  int Read_(Tcl_Interp *interp, char *formatString, 
	    Tk_PhotoHandle imageHandle, int destX, int destY,
	    int width, int height, int srcX, int srcY);
  int DoExtension _ANSI_ARGS_((int label,
					 int *transparent));
  int GetCode _ANSI_ARGS_((int code_size, int flag));
  int GetDataBlock _ANSI_ARGS_((unsigned char *buf));
  int LWZReadByte _ANSI_ARGS_((int flag, int input_code_size));
  int ReadColorMap _ANSI_ARGS_((int number, 
				unsigned char buffer[3][MAXCOLORMAPSIZE]));
  int ReadGIFHeader _ANSI_ARGS_((int *widthPtr, int *heightPtr));
  int ReadImage _ANSI_ARGS_((Tcl_Interp *interp, char *imagePtr, 
			     int len, int height,
			     unsigned char cmap[3][MAXCOLORMAPSIZE],
			     int interlace, int transparent));

  int ReadData(void *ptr, size_t size, size_t nmemb);
};


static class GIFClass : public TclClass {
public:
  GIFClass() : TclClass("Image/GIF") { };
  TclObject *create(int argc, const char*const* argv) {
    // argv[4] onwards are the real args
    argv += 4;
    argc -= 4;

    if (argc!=1) {
      Tcl_Channel errChannel = Tcl_GetStdChannel(TCL_STDERR);
      if (errChannel) {
	Tcl_Write(errChannel, 
		  "Invalid number of args while creating Image/GIF\n", -1);
      }
      return NULL;
    }

    return new GIFObject(argv[0]);
  }
  
} gif_class_object;




Tk_PhotoImageFormat tkImgObjGIF = {
  "GIF-TclObject",  /* name */
  NULL,             /* fileMatchProc */
  GIFObject::Match, /* stringMatchProc */
  NULL,             /* fileReadProc */
  GIFObject::Read,  /* stringReadProc */
  NULL,             /* fileWriteProc */
  NULL,             /* stringWriteProc */
};



GIFObject::GIFObject(const char *string)
  : buffer(NULL), size(0), nextByte(NULL)
{
  char *ptr;
  int maxBytes = strlen(string);
  buffer = new char [maxBytes+1];
  if (buffer==NULL) return;

  ptr = buffer;
  while (*string!='\0') {
    if (*string=='\\' && *(string+1)=='#') {
      *ptr++ = '\0';
      string += 2;
    }
    else if (*string=='\\' && *(string+1)=='\\') {
      *ptr++ = '\\';
      string += 2;
    }
    else
      *ptr++ = *string++;
  }

  size = ptr-buffer;
  nextByte = buffer;
}


GIFObject::~GIFObject()
{
  if (buffer!=NULL) delete [] buffer;
  buffer = NULL;
  size = 0;
  nextByte = NULL;
}


int
GIFObject::ReadData(void *ptr, size_t sizeOfMemb, size_t nmemb)
{
  size_t left, i;
  char *cptr = (char*) ptr;

  if (sizeOfMemb==0) {
    return 0;
  }

  left = nmemb;
  while (left > 0) {
    if (nextByte + sizeOfMemb > buffer + size) {
      // out of data
      return nmemb - left;
    }
    for (i=0; i<sizeOfMemb; i++)
      *cptr++ = *nextByte++;
    left--;
  }

  return nmemb;
}





/*
 *----------------------------------------------------------------------
 *
 * GIFObject::Match --
 *
 *  This procedure is invoked by the photo image type to see if
 *  a file contains image data in GIF format.
 *
 * Results:
 *  The return value is 1 if the first characters in object o look
 *  like GIF data, and 0 otherwise.
 *
 * Side effects:
 *  The access position in o may change.
 *
 *----------------------------------------------------------------------
 */

int
GIFObject::Match(char *objectName, char *formatString, 
		 int *widthPtr, int *heightPtr)
{
  GIFObject *o = (GIFObject*) Tcl::instance().lookup(objectName);
  if (o==NULL) {
    return 0;
  }
  o->Reset();
  return o->ReadGIFHeader(widthPtr, heightPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * GIFObject::Read --
 *
 *	This procedure is called by the photo image type to read
 *	GIF format data from a TclObject and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed, and new data is
 *	added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

int
GIFObject::Read(Tcl_Interp *interp, char *objectName, char *formatString, 
		Tk_PhotoHandle imageHandle, int destX, int destY,
		int width, int height, int srcX, int srcY)
{
  GIFObject *o = (GIFObject*) Tcl::instance().lookup(objectName);
  if (o==NULL) {
    Tcl_AppendResult(interp, "Invalid object name ",
		     (objectName==NULL) ? "(nil)" : objectName, NULL);
    return TCL_ERROR;
  }
  o->Reset();
  return o->Read_(interp, formatString, imageHandle, destX, destY, width, 
		  height, srcX, srcY);
}  


/*
 *----------------------------------------------------------------------
 *
 * GIFObject::Read_ --
 *
 *	This procedure is called by the photo image type to read
 *	GIF format data from a TclObject and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed, and new data is
 *	added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

int
GIFObject::Read_(Tcl_Interp *interp, char *formatString, 
		 Tk_PhotoHandle imageHandle, int destX, int destY,
		 int width, int height, int srcX, int srcY)
{
  int fileWidth, fileHeight;
  int nBytes;
  Tk_PhotoImageBlock block;
  unsigned char buf[100];
  int bitPixel;
  unsigned int colorResolution;
  unsigned int background;
  unsigned int aspectRatio;
  unsigned char localColorMap[3][MAXCOLORMAPSIZE];
  unsigned char colorMap[3][MAXCOLORMAPSIZE];
  int useGlobalColormap;
  int transparent = -1;

  if (!ReadGIFHeader(&fileWidth, &fileHeight)) {
    Tcl_AppendResult(interp, "couldn't read GIF header from TclObject \"",
		     (name()==NULL) ? "(nil)":name(), "\"", NULL);
    return TCL_ERROR;
  }
  if ((fileWidth <= 0) || (fileHeight <= 0)) {
    Tcl_AppendResult(interp, "GIF image object \"", 
		     (name()==NULL) ? "(nil)":name(),
		     "\" has dimension(s) <= 0", (char *) NULL);
    return TCL_ERROR;
  }

  if (ReadData(buf, 1, 3) != 3) {
    return TCL_OK;
  }
  bitPixel = 2<<(buf[0]&0x07);
  colorResolution = (((buf[0]&0x70)>>3)+1);
  background = buf[1];
  aspectRatio = buf[2];

  if (BitSet(buf[0], LOCALCOLORMAP)) {    /* Global Colormap */
    if (!ReadColorMap(bitPixel, colorMap)) {
      Tcl_AppendResult(interp, "error reading color map",
		       (char *) NULL);
      return TCL_ERROR;
    }
  }

  if ((srcX + width) > fileWidth) {
    width = fileWidth - srcX;
  }
  if ((srcY + height) > fileHeight) {
    height = fileHeight - srcY;
  }
  if ((width <= 0) || (height <= 0)
      || (srcX >= fileWidth) || (srcY >= fileHeight)) {
    return TCL_OK;
  }

  Tk_PhotoExpand(imageHandle, destX + width, destY + height);

  block.width = fileWidth;
  block.height = fileHeight;
  block.pixelSize = 3;
  block.pitch = 3 * fileWidth;
  block.offset[0] = 0;
  block.offset[1] = 1;
  block.offset[2] = 2;
  nBytes = fileHeight * block.pitch;
  block.pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);

  while (1) {
    if (ReadData(buf, 1, 1) != 1) {
      /*
       * Premature end of image.  We should really notify
       * the user, but for now just show garbage.
       */

      break;
    }

    if (buf[0] == ';') {
      /*
       * GIF terminator.
       */

      break;
    }

    if (buf[0] == '!') {
      /*
       * This is a GIF extension.
       */

      if (ReadData(buf, 1, 1) != 1) {
	interp->result =
	  "error reading extension function code in GIF image";
	goto error;
      }
      if (DoExtension(buf[0], &transparent) < 0) {
	interp->result = "error reading extension in GIF image";
	goto error;
      }
      continue;
    }

    if (buf[0] != ',') {
      /*
       * Not a valid start character; ignore it.
       */
      continue;
    }

    if (ReadData(buf, 1, 9) != 9) {
      interp->result = "couldn't read left/top/width/height in GIF image";
      goto error;
    }

    useGlobalColormap = ! BitSet(buf[8], LOCALCOLORMAP);

    bitPixel = 1<<((buf[8]&0x07)+1);

    if (!useGlobalColormap) {
      if (!ReadColorMap(bitPixel, localColorMap)) {
	Tcl_AppendResult(interp, "error reading color map", 
			 (char *) NULL);
	goto error;
      }
      if (ReadImage(interp, (char *) block.pixelPtr, fileWidth,
		    fileHeight, localColorMap, BitSet(buf[8], INTERLACE),
		    transparent) != TCL_OK) {
	goto error;
      }
    } else {
      if (ReadImage(interp, (char *) block.pixelPtr, fileWidth,
		    fileHeight, colorMap, BitSet(buf[8], INTERLACE),
		    transparent) != TCL_OK) {
	goto error;
      }
    }
    break;
  }

  Tk_PhotoPutBlock(imageHandle, &block, destX, destY, fileWidth, fileHeight);
  ckfree((char *) block.pixelPtr);
  return TCL_OK;

error:
  ckfree((char *) block.pixelPtr);
  return TCL_ERROR;

}

/*
 *----------------------------------------------------------------------
 *
 * ReadGIFHeader --
 *
 *	This procedure reads the GIF header from the beginning of a
 *	GIF file and returns the dimensions of the image.
 *
 * Results:
 *	The return value is 1 if file "f" appears to start with
 *	a valid GIF header, 0 otherwise.  If the header is valid,
 *	then *widthPtr and *heightPtr are modified to hold the
 *	dimensions of the image.
 *
 * Side effects:
 *	The access position in f advances.
 *
 *----------------------------------------------------------------------
 */

int
GIFObject::ReadGIFHeader(int *widthPtr, int *heightPtr)
{
  unsigned char buf[7];

  if ((ReadData(buf, 1, 6) != 6)
      || ((strncmp("GIF87a", (char *) buf, 6) != 0)
	  && (strncmp("GIF89a", (char *) buf, 6) != 0))) {
    return 0;
  }

  if (ReadData(buf, 1, 4) != 4) {
    return 0;
  }

  *widthPtr = LM_to_uint(buf[0],buf[1]);
  *heightPtr = LM_to_uint(buf[2],buf[3]);
  return 1;
}

/*
 *-----------------------------------------------------------------
 * The code below is copied from the giftoppm program and modified
 * just slightly.
 *-----------------------------------------------------------------
 */

int
GIFObject::ReadColorMap(int number, unsigned char buffer[3][MAXCOLORMAPSIZE])
{
  int     i;
  unsigned char   rgb[3];

  for (i = 0; i < number; ++i) {
    if (! ReadOK(rgb, sizeof(rgb)))
      return 0;

    buffer[CM_RED][i] = rgb[0] ;
    buffer[CM_GREEN][i] = rgb[1] ;
    buffer[CM_BLUE][i] = rgb[2] ;
  }
  return 1;
}



int
GIFObject::DoExtension(int label, int *transparent)
{
  static unsigned char buf[256];
  int count = 0;

  switch (label) {
  case 0x01:      /* Plain Text Extension */
    break;

  case 0xff:      /* Application Extension */
    break;

  case 0xfe:      /* Comment Extension */
    do {
      count = GetDataBlock((unsigned char*) buf);
    } while (count > 0);
    return count;

  case 0xf9:      /* Graphic Control Extension */
    count = GetDataBlock((unsigned char*) buf);
    if (count < 0) {
      return 1;
    }
    if ((buf[0] & 0x1) != 0) {
      *transparent = buf[3];
    }

    do {
      count = GetDataBlock((unsigned char*) buf);
    } while (count > 0);
    return count;
  }

  do {
    count = GetDataBlock((unsigned char*) buf);
  } while (count > 0);
  return count;
}

static int ZeroDataBlock = 0;

int
GIFObject::GetDataBlock(unsigned char *buf)
{
  unsigned char   count;

  if (! ReadOK(&count,1)) {
    return -1;
  }

  ZeroDataBlock = count == 0;

  if ((count != 0) && (! ReadOK(buf, count))) {
    return -1;
  }

  return count;
}


int
GIFObject::ReadImage(Tcl_Interp *interp, char *imagePtr, int len, int height, 
		     unsigned char cmap[3][MAXCOLORMAPSIZE], 
		     int interlace, int transparent)
{
  unsigned char   c;
  int     v;
  int     xpos = 0, ypos = 0, pass = 0;
  char 	*colStr;


  /*
   *  Initialize the Compression routines
   */
  if (! ReadOK(&c,1))  {
    Tcl_AppendResult(interp, "error reading GIF image: ",
		     Tcl_PosixError(interp), (char *) NULL);
    return TCL_ERROR;
  }

  if (LWZReadByte(1, c) < 0) {
    interp->result = "format error in GIF image";
    return TCL_ERROR;
  }

  if (transparent != -1) {
    colStr = Tcl_GetVar(interp, "TRANSPARENT_GIF_COLOR", 0L);
    if (colStr != NULL) {
      XColor *colorPtr;
      colorPtr = Tk_GetColor(interp, Tk_MainWindow(interp), 
			     Tk_GetUid(colStr));
      if (colorPtr) {
	cmap[CM_RED][transparent] = colorPtr->red >> 8;
	cmap[CM_GREEN][transparent] = colorPtr->green >> 8;
	cmap[CM_BLUE][transparent] = colorPtr->blue >> 8;
	Tk_FreeColor(colorPtr);
      }
    }
  }

  while ((v = LWZReadByte(0,c)) >= 0 ) {

    imagePtr[ (xpos*3)  +  (ypos *len*3)] = cmap[CM_RED][v];
    imagePtr[ (xpos*3)  +  (ypos *len*3) +1] = cmap[CM_GREEN][v];
    imagePtr[ (xpos*3)  +  (ypos *len*3) +2] = cmap[CM_BLUE][v];

    ++xpos;
    if (xpos == len) {
      xpos = 0;
      if (interlace) {
	switch (pass) {
	case 0:
	case 1:
	  ypos += 8; break;
	case 2:
	  ypos += 4; break;
	case 3:
	  ypos += 2; break;
	}

	if (ypos >= height) {
	  ++pass;
	  switch (pass) {
	  case 1:
	    ypos = 4; break;
	  case 2:
	    ypos = 2; break;
	  case 3:
	    ypos = 1; break;
	  default:
	    return TCL_OK;
	  }
	}
      } else {
	++ypos;
      }
    }
    if (ypos >= height)
      break;
  }
  return TCL_OK;
}

int
GIFObject::LWZReadByte(int flag, int input_code_size)
{
  static int  fresh = 0;
  int     code, incode;
  static int  code_size, set_code_size;
  static int  max_code, max_code_size;
  static int  firstcode, oldcode;
  static int  clear_code, end_code;
  static int  table[2][(1<< MAX_LWZ_BITS)];
  static int  stack[(1<<(MAX_LWZ_BITS))*2], *sp;
  register int    i;


  if (flag) {

    set_code_size = input_code_size;
    code_size = set_code_size+1;
    clear_code = 1 << set_code_size ;
    end_code = clear_code + 1;
    max_code_size = 2*clear_code;
    max_code = clear_code+2;

    GetCode(0, 1);

    fresh = 1;

    for (i = 0; i < clear_code; ++i) {
      table[0][i] = 0;
      table[1][i] = i;
    }
    for (; i < (1<<MAX_LWZ_BITS); ++i) {
      table[0][i] = table[1][0] = 0;
    }

    sp = stack;

    return 0;

  } else if (fresh) {

    fresh = 0;
    do {
      firstcode = oldcode = GetCode(code_size, 0);
    } while (firstcode == clear_code);
    return firstcode;
  }

  if (sp > stack)
    return *--sp;

  while ((code = GetCode(code_size, 0)) >= 0) {
    if (code == clear_code) {
      for (i = 0; i < clear_code; ++i) {
	table[0][i] = 0;
	table[1][i] = i;
      }

      for (; i < (1<<MAX_LWZ_BITS); ++i) {
	table[0][i] = table[1][i] = 0;
      }

      code_size = set_code_size+1;
      max_code_size = 2*clear_code;
      max_code = clear_code+2;
      sp = stack;
      firstcode = oldcode = GetCode(code_size, 0);
      return firstcode;

    } else if (code == end_code) {
      int     count;
      unsigned char   buf[260];

      if (ZeroDataBlock)
	return -2;

      while ((count = GetDataBlock(buf)) > 0)
	;

      if (count != 0)
	return -2;
    }

    incode = code;

    if (code >= max_code) {
      *sp++ = firstcode;
      code = oldcode;
    }

    while (code >= clear_code) {
      *sp++ = table[1][code];
      if (code == table[0][code]) {
	return -2;

	/*
	 * Used to be this instead, Steve Ball suggested
	 * the change to just return.

	 printf("circular table entry BIG ERROR\n");
	 */
      }
      code = table[0][code];
    }

    *sp++ = firstcode = table[1][code];

    if ((code = max_code) <(1<<MAX_LWZ_BITS)) {

      table[0][code] = oldcode;
      table[1][code] = firstcode;
      ++max_code;
      if ((max_code>=max_code_size) && (max_code_size < (1<<MAX_LWZ_BITS))) {
	max_code_size *= 2;
	++code_size;
      }
    }

    oldcode = incode;

    if (sp > stack)
      return *--sp;
  }
  return code;
}


int
GIFObject::GetCode(int code_size, int flag)
{
  static unsigned char    buf[280];
  static int      curbit, lastbit, done, last_byte;
  int             i, j, ret;
  unsigned char   count;

  if (flag) {
    curbit = 0;
    lastbit = 0;
    done = 0;
    return 0;
  }


  if ( (curbit+code_size) >= lastbit) {
    if (done) {
      /* ran off the end of my bits */
      return -1;
    }
    buf[0] = buf[last_byte-2];
    buf[1] = buf[last_byte-1];

    if ((count = GetDataBlock(&buf[2])) == 0)
      done = 1;

    last_byte = 2 + count;
    curbit = (curbit - lastbit) + 16;
    lastbit = (2+count)*8 ;
  }

  ret = 0;
  for (i = curbit, j = 0; j < code_size; ++i, ++j)
    ret |= ((buf[ i / 8 ] & (1 << (i % 8))) != 0) << j;


  curbit += code_size;

  return ret;
}
