# This file contains the Dali subprogram base class.
# The DaliSubprogram object responds to graph interface communication
# recieved through its recv_graph_comm method and 
# sends graph interface communication through the send_graph_comm
# method. 

# Object fields
#
#   input_id_list_    : list of input ids
#   output_id_list_   : list of output ids
#   input_info_       : array of info on input_ids. For each input id $i,
#         ($i,spec)       : multicast spec for input data
#         ($i,trigger)    : 0/1 indicating if input is execution trigger
#         ($i,buffertype) : indicates what type of buffer this Dali subprogram
#                           expects the input to be dumped into
#         ($i,buffername) : indicates the name of the buffer object of the
#                           type indicated above
#         ($i,decoder)    : decoder object for input
#   output_id_list_ of output ids
#   output_info_      : array of info on output_ids. For each output id $o,
#         ($o,spec)      : multicast spec of where output goes
#         ($o,buffertype): indicates type of above buffer
#         ($o,encoder)   : name of encoder object.
#         ($o,format)    : format of output
#         ($o,vagent)    : name of video agent used as a transmitter
#         ($o,bp)        : name of buffer pool
#   parameter_id_list_: list of parameter ids
#   parameter_info_   : array on info on parameters. For each parameter $p,
#         ($p,oname)     : name of parameter object.
#   vagent_array_        : array associated by "addr,port" indexes that holds
#                          the name of video agents used by this object.
#   tvector_             : list of pairs (id,ts) indicating that trigger is
#                          waiting for frame with timestamp ts to appear on
#                          on id.

# Object methods
#   init : initialization
#   recv_graph_comm : handles incoming graph interface communication
#   send_graph_comm : handles outgoing graph interface communication
#   set_input_spec : handles changes to input spec
#   set_output_spec : handles changes to output spec
#   set_output_format : handles changes to output format
#   trigger : executes Dali code to produce 1 frame of output
#   set_output_geometry : handles changes to output geometry

import PsvpVideoAgent
import AnnounceListenManager/GraphComm

Class DaliSubprogram

DaliSubprogram instproc init {id control_spec} {
    $self next

    $self instvar input_id_list_;
    $self instvar input_info_;
    $self instvar output_id_list_;
    $self instvar output_info_;
    $self instvar parameter_id_list_;
    $self instvar parameter_info_;
    $self instvar vagent_array_;
    $self instvar comm_obj_;
    $self instvar id_;

    set input_id_list_ "";
    set output_id_list_ "";
    set parameter_id_list_ "";

    set id_ $id;
    set comm_obj_ [new AnnounceListenManager/GraphComm [$self get_option mtu] $control_spec];
    $comm_obj_ register $id "$self recv_graph_comm ";
}

DaliSubprogram instproc recv_graph_comm {rid mesg} {
    $self instvar input_id_list_ input_info_
    $self instvar output_id_list_ output_info_
    $self instvar parameter_id_list_ parameter_info_

    set cmd [lindex $mesg 0];

    switch -exact -- $cmd {
	query_input_ids {
	    $self send_graph_comm $rid [list input_ids $input_id_list_];
	}
	query_input_info {
	    set in_id [lindex $mesg 1];
	    if {[lsearch $input_id_list_ $in_id] != -1} {
		$self send_graph_comm $rid [list input_info $in_id [list \
			[list spec $input_info_($in_id,spec)] \
			[list trigger $input_info_($in_id,trigger)]]];
	    } else {
		$self send_graph_comm $rid [list input_info $in_id [list \
			[list error "No such input_id $in_id"]]];
	    }
	}
	query_output_ids {
	    $self send_graph_comm $rid [list output_ids $output_id_list_];
	}
	query_output_info {
	    set out_id [lindex $mesg 1];
	    if {[lsearch $output_id_list_ $out_id] != -1} {
		$self send_graph_comm $rid [list output_info $out_id [list \
			[list spec $output_info_($out_id,spec)]]];
	    } else {
		$self send_graph_comm $rid [list output_info $out_id [list \
			[list error "No such output_id $out_id"]]];
	    }
	}
	query_parameter_ids {
	    $self send_graph_comm $rid [list parameter_ids $parameter_id_list_];
	}
	query_parameter_info {
	    set p_id [lindex $mesg 1];
	    if {[lsearch $parameter_id_list_ $p_id] != -1} {
		set pobj $parameter_info_($p_id,oname);

		$self send_graph_comm $rid [list parameter_info $p_id [list \
			[list value [$pobj get]] \
			[list type [$pobj type]] \
			[list domain [$pobj domain]]]];
	    } else {
		$self send_graph_comm $rid [list parameter_info $p_id [list \
			[list error "No such parameter_id $p_id"]]];
	    }
	}
	set_input_spec {
	    set in_id [lindex $mesg 1];
	    set spec [lindex $mesg 2];
	    if {[lsearch $input_id_list_ $in_id] != -1} {
		$self set_input_spec $in_id $spec;
	    }
	}
	set_output_spec {
	    set out_id [lindex $mesg 1];
	    set spec [lindex $mesg 2];
	    if {[lsearch $output_id_list_ $out_id] != -1} {
		$self set_output_spec $out_id $spec;
	    }
	}
	set_output_format {
	    set out_id [lindex $mesg 1];
	    set format [lindex $mesg 2];
	    if {[lsearch $output_id_list_ $out_id] != -1} {
		$self set_output_format $out_id $format;
	    }
	}
	set_input_trigger {
	    set in_id [lindex $mesg 1];
	    set trigger_flag [lindex $mesg 2];
	    if {[lsearch $input_id_list_ $in_id] != -1} {
		if {$trigger_flag != $input_info_($in_id,trigger)} {
		    set input_info_($in_id,trigger) $trigger_flag;

		    if {$input_info_($in_id,decoder) != ""} {
			if {$input_info_($in_id,trigger) != 0} {
			    $input_info_($in_id,decoder) set_callback "$self trigger";
			} else {
			    $input_info_($in_id,decoder) set_callback "";
			} 
		    }
		}
	    }
	}
	set_rate_limited_input_trigger {
	    set in_id [lindex $mesg 1];
	    set trigger_flag [lindex $mesg 2];
	    if {$trigger_flag == 1} {
		set trigger_offset [lindex $mesg 3];
		set trigger_period [lindex $mesg 4];
	    } else {
		set trigger_offset 0;
		set trigger_period 0;
	    }

	    if {[lsearch $input_id_list_ $in_id] != -1} {
		set input_info_($in_id,rate_trigger_offset) $trigger_offset;
		set input_info_($in_id,rate_trigger_period) $trigger_period;

		if {$trigger_flag != $input_info_($in_id,trigger)} {
		    set input_info_($in_id,trigger) $trigger_flag;
		    
		    if {$input_info_($in_id,decoder) != ""} {
			if {$input_info_($in_id,trigger) != 0} {
			    $input_info_($in_id,decoder) set_callback "$self rated_trigger $in_id";
			} else {
			    $input_info_($in_id,decoder) set_callback "";
			}
		    }
		}
	    }
	}
	trigger {
	    $self trigger;
	}
	trigger_vector {
	    set vector [lindex $mesg 1];
	    $self trigger_vector $vector;
	}
	set_output_geometry {
	    if {[llength $mesg] == 3} {
		set out_id [lindex $mesg 1];
		set geometry [lindex $mesg 2];
		if {[lsearch $output_id_list_ $out_id] != -1} {
		    $self set_output_geometry $out_id $geometry;
		}
	    } else {
		set geometry [lindex $mesg 1];
		foreach out_id $output_id_list_ {
		    $self set_output_geometry $out_id $geometry;
		}
	    }
	}
	set_parameter {
	    set p_id [lindex $mesg 1]
	    set p_value [lindex $mesg 2];

	    if {[lsearch $parameter_id_list_ $p_id] != -1} {
		set pobj $parameter_info_($p_id,oname);
		if {[$pobj get] != $p_value} {
		    $pobj set $p_value;
		}
	    }
	}
    }
}

DaliSubprogram instproc send_debug_mesg {mesg} {
    $self send_graph_comm -1 "debug $mesg";
}

DaliSubprogram instproc send_completion_token {} {
    $self send_graph_comm -1 "trigger_completion_token";
}

DaliSubprogram instproc send_graph_comm {rid mesg} {
    $self instvar comm_obj_;
    $self instvar id_;

    $comm_obj_ do_announce $id_ $rid $mesg;
}

DaliSubprogram instproc set_input_spec {id spec} {
    $self instvar input_info_
    $self instvar vagent_array_;

    # If alread set properly, ignore.

    if {$input_info_($id,spec) == $spec} {
	return;
    }

    # If input already associated with a spec, delete
    # decoder no longer needed.

    if {$input_info_($id,spec) != ""} {
	set decoder $input_info_($id,decoder);
	if {$decoder != ""} {
	    set vagent [$decoder set agent_];
	    $vagent delete_decoder $decoder;
	    set input_info_($id,decoder) "";
	}	    
	set input_info_($id,spec) "";
    }

    set spec_split [split $spec "/"];
	
    set addr [lindex $spec_split 0];
    set port [lindex $spec_split 1];
    set srcid [lindex $spec_split 2];
    
    # Find video agent handling this addr/port creating
    # one if none exists.

    if {![info exists vagent_array_($addr,$port)]} {
	set vagent_array_($addr,$port) [new PsvpVideoAgent $addr/$port];
    }
    set vagent $vagent_array_($addr,$port);

    # Retrieve source associated with target srcid
	
    set src [$vagent get_source_by_id $srcid];

    # If no source yet, establish callback with video
    # agent for when source does appear

    if {$src == ""} {
	$vagent set_create_decoder_callback $srcid "$self set_input_spec_cb $id $spec";
    } else {

	# Create appropriate decoder for input source format and
	# target buffer type.

	set fmt_name [$vagent classmap [$src format_name]];

	set btype $input_info_($id,buffertype);

	set bname $input_info_($id,buffername);

	set decoder [new Module/VideoDecoder/${fmt_name}To${btype}];

	if {$decoder == ""} {
	    # No such decoder. Create a NULL decoder and associate that.
	    set decoder [new Module/VideoDecoder/Null];
	} else {
	    $decoder set_frame_buffer $bname;
	    if {$input_info_($id,trigger) != 0} {
		$decoder set_callback "$self trigger";
	    }
	}

	# Instruct video agent to set up decoder as target for source

	$vagent set_src_decoder $src $decoder;
	
	# Update input info array

	set input_info_($id,decoder) $decoder;
    }
    set input_info_($id,spec) $spec;
}

DaliSubprogram instproc set_input_spec_cb {id spec src} {
    $self instvar input_info_
    $self instvar vagent_array_

    # Check to make sure spec is still good.

    if {$input_info_($id,spec) != $spec} {
	puts "$input_info_($id,spec) != $spec !!!";
	return;
    }

    if {$input_info_($id,decoder) != ""} {
	puts "$input_info_($id,decoder) != {} !!!";
	return;
    }

    # Find videoagent and source

    set spec_split [split $spec "/"];
	
    set addr [lindex $spec_split 0];
    set port [lindex $spec_split 1];
    set srcid [lindex $spec_split 2];

    set vagent $vagent_array_($addr,$port);

#    set src [$vagent get_source_by_id $srcid];

#    if {$src == ""} {
#	error "Video agent callback called when source not actually available";
#    }

    # Setup decoder.
    
    set fmt_name [$vagent classmap [$src format_name]];

    set btype $input_info_($id,buffertype);

    set bname $input_info_($id,buffername);

    set decoder [new Module/VideoDecoder/${fmt_name}To${btype}];

    if {$decoder == ""} {
	# No such decoder. Create a NULL decoder and associate that.
	set decoder [new Module/VideoDecoder/Null];
    } else {
	$decoder set_frame_buffer $bname;
	if {$input_info_($id,trigger) != 0} {
	    $decoder set_callback "$self trigger";
	}
    }

    # Update input info array

    set input_info_($id,decoder) $decoder;

    return $decoder;
}

DaliSubprogram instproc set_output_spec {id spec} {
    $self instvar output_info_
    $self instvar vagent_array_

    # If already set properly, ignore.

    if {$output_info_($id,spec) == $spec} {
	return;
    }

    # If out spec already set, unset it.

    if {$output_info_($id,spec) != ""} {
	set encoder $output_info_($id,encoder);
	if {$encoder != ""} {
	    delete $encoder;
	}
	set bp $output_info_($id,bp);
	if {$bp != ""} {
	    delete $bp;
	}
	set output_info_($id,encoder) "";
	set output_info_($id,spec) "";
	set output_info_($id,vagent) "";
	set output_info_($id,bp) "";
    }

    # Find video agent for addr/port

    set spec_split [split $spec "/"];
	
    set addr [lindex $spec_split 0];
    set port [lindex $spec_split 1];

    if {![info exists vagent_array_($addr,$port)]} {
	set vagent_array_($addr,$port) [new PsvpVideoAgent $addr/$port];
    }
    set vagent $vagent_array_($addr,$port);

    # Create encoder for buffertype and target format

    set format $output_info_($id,format);
    set btype $output_info_($id,buffertype);

    set encoder [new Module/VideoEncoder/${btype}To${format}];

    # Set video agent transmitter as encoder target

    set bp "";

    if {$encoder != ""} {
	set ssrc [$vagent get_local_srcid];
	set bp [new BufferPool/RTP];
	$bp srcid $ssrc;
	
	$encoder buffer-pool $bp;
	$encoder mtu [$self get_option mtu];
	$encoder target [$vagent get_transmitter];

	# Total hack XXXXXXXX
	$vagent local_bandwidth 30000000
    }

    # Update variables
    
    set output_info_($id,spec) $spec;
    set output_info_($id,encoder) $encoder;
    set output_info_($id,vagent) $vagent;
    set output_info_($id,bp) $bp;
}

DaliSubprogram instproc set_output_format {id format} {
    $self instvar output_info_

    if {$output_info_($id,format) == $format} {
	return;
    }

    set output_info_($id,format) $format;

    if {$output_info_($id,spec) == ""} {
	return;
    }
    
    if {$output_info_($id,encoder) != ""} {
	delete $output_info_($id,encoder);
    }
    set output_info_($id,encoder) "";

    if {$output_info_($id,bp) != ""} {
	delete $output_info_($id,bp);
    }
    set output_info_($id,bp) "";

    set btype $output_info_($id,buffertype);

    set encoder [new Module/VideoEncoder/${btype}To${format}];
    set bp "";

    if {$encoder != ""} {
	set ssrc [$output_info_($id,vagent) get_local_srcid];
	set bp [new BufferPool/RTP];
	$bp srcid $ssrc;
	
	$encoder buffer-pool $bp;
	$encoder mtu [$self get_option mtu];
	$encoder target [$output_info_($id,vagent) get_transmitter]

	# Total hack XXXXXXXX
	$vagent local_bandwidth 30000000
    }
    
    set output_info_($id,encoder) $encoder;
    set output_info_($id,bp) $bp;

}

DaliSubprogram instproc trigger_vector {vector} {
    $self instvar id_;
    $self instvar tvector_ input_info_ input_id_list_;
    $self instvar tvector_queue_;

    #    puts "$id_ :: $vector";
    
    if {[info exists tvector_]} {
	# Got called while already processing a trigger 
	# vector. For now, send debug message and do
	# nothing.
	
	$self send_debug_mesg "Trigger_vector queued";

#	lappend tvector_queue_ $vector;
	set tvector_queue_ [list $vector];

	return;
    }
    
    set tvector_ "";
    
    foreach entry $vector {
	set id [lindex $entry 0];
	set ts [lindex $entry 1];
	
	set cur_ts [$input_info_($id,buffername) set ts_];
	if {$cur_ts > $ts} {
	    # Missed
	    foreach id $input_id_list_ {
		if {$input_info_($id,decoder) != "" } {
		    $input_info_($id,decoder) set_callback "";
		    $input_info_($id,decoder) delay_off;
		}
	    }
	    unset tvector_;
	    $self send_debug_mesg "Missed timestamp for input $id";
	    $self check_tvector_queue
	    return;
	} elseif {$cur_ts == $ts} {
	    # Right on, set delay.
	    if {$input_info_($id,decoder) != "" } {
		$input_info_($id,decoder) delay_on;
	    }
	} else {
	    # Waiting.
	    if {$input_info_($id,decoder) == ""} {
		# Big problem!!!
		foreach id $input_id_list_ {
		    if {$input_info_($id,decoder) != "" } {
			$input_info_($id,decoder) set_callback "";
			$input_info_($id,decoder) delay_off;
		    }
		}
		$self send_debug_mesg "No decoder for input $id";
		unset tvector_;
		$self check_tvector_queue
		return;
	    } else {
		# Set callback
		$input_info_($id,decoder) set_callback "$self check_vector $id $ts";
		lappend tvector_ [list $id $ts];
	    }
	}
    }
    if {$tvector_ == ""} {
	# Ready to go.
	$self trigger;
	foreach id $input_id_list_ {
	    if {$input_info_($id,decoder) != "" } {
		$input_info_($id,decoder) set_callback "";
		$input_info_($id,decoder) delay_off;
	    }
	}
	unset tvector_;
	$self check_tvector_queue
    }
}

DaliSubprogram instproc check_tvector_queue {} {
    $self instvar tvector_queue_;

    if {[info exists tvector_queue_]} {
	if {[llength $tvector_queue_] == 0} {
	    unset tvector_queue_;
	} else {
	    set next_tvector [lindex $tvector_queue_ 0];
	    set tvector_queue_ [lrange $tvector_queue_ 1 end];
	    $self send_debug_mesg "Processing trigger vector from queue";
	    $self trigger_vector $next_tvector;
	    return;
	}
    }
}
    
DaliSubprogram instproc check_vector {id ts} {
    $self instvar input_info_ tvector_ input_id_list_

    set cur_ts [$input_info_($id,buffername) set ts_];

    if {$cur_ts > $ts} {
	# Missed
	foreach id $input_id_list_ {
	    if {$input_info_($id,decoder) != "" } {
		$input_info_($id,decoder) set_callback "";
		$input_info_($id,decoder) delay_off;
	    }
	}
	unset tvector_;
	$self send_debug_mesg "Check vector missed on input $id";
	$self check_tvector_queue;
	return;

    } elseif {$cur_ts == $ts} {
	# Right on
	set new_tvector "";
	foreach entry $tvector_ {
	    set eid [lindex $entry 0];
	    if {$eid != $id} {
		lappend new_tvector $entry;
	    }
	}
	set tvector_ $new_tvector;
	
	$input_info_($id,decoder) set_callback "";
	$input_info_($id,decoder) delay_on;

	if {$tvector_ == ""} {
	    # Ready to go.
	    $self trigger;
	    foreach id $input_id_list_ {
		if {$input_info_($id,decoder) != "" } {
		    $input_info_($id,decoder) set_callback "";
		    $input_info_($id,decoder) delay_off;
		}
	    }
	    unset tvector_;
	    $self check_tvector_queue;
	}
    }
}

DaliSubprogram instproc rated_trigger {in_id} {
    $self instvar input_info_;
    
    set period $input_info_($in_id,rate_trigger_period);
    set offset $input_info_($in_id,rate_trigger_offset);
    
    set bname $input_info_($in_id,buffername);

    set ts [$bname set ts_];
    
    if {$ts > $offset} {
#	puts "$ts > $offset";
	$self trigger;
	set input_info_($in_id,rate_trigger_offset) [expr (int(($ts - $offset) / $period) + 1) * $period + $offset];
    } else {
#	puts "$ts < $offset";
    }
}


DaliSubprogram instproc trigger {} {
    puts "No default trigger action";
}

DaliSubprogram instproc set_output_geometry {out_id geometry} {
    puts "No default output geometry action";
}




