# ---------------------------------------------------------------------------
# This software is in the public domain, furnished "as is", without technical
# support, and with no warranty, express or implied, as to its usefulness for
# any purpose.
#
# tips.tcl
# Popup tips for Tk
# ---------------------------------------------------------------------------

# -- module -----------------------------------------------------------------
#
# This package provides popup tips for widgets in Tcl/Tk.
#
# Simply prefix your widget creation statement with TipWidget and use a new
# option "-tip msg" while creating the widget
#   e.g. TipWidget button .b -text "Foo" -tip "Bar"
# ---------------------------------------------------------------------------


WidgetClass TipWidget -configspec {
	{ -tip tip Tip {foobar} config_tip cget_tip }
} -default {
	{ *background  LemonChiffon }
	{ *forground   Black }
	{ *relief      raised }
	{ *borderWidth 1 }
	{ *justify     center }
	{ *font        "-*-helvetica-medium-r-normal-*-*-120-*-*-*-*-*-*" }
	{ *windowdelay 500 }
	{ *yoffset     5 }
	{ *xoffset     20 }
}


# -- proc ------------------------------------------------------------------
# create
#
# The widget creation method accepts the widget class (button, entry, etc.)
# as the first argument and the widget name as the second. But the default
# "WidgetClass create" method expects the widget name as the first argument,
# so flip ther two args here
# --------------------------------------------------------------------------
TipWidget proc create { cl widget args } {
	eval [list $self] next [list $widget] [list $cl] $args
}


# -- proc -------------------------------------------------------------------
# init_
#
# Initialize the tips.  Currently, this just means adding X resources
# and setting up bindings.  Override these in your own X resources, if
# you must.
# ---------------------------------------------------------------------------
TipWidget proc init_ {} {
	$self instvar enabled_ currently_tipping_
	set enabled_ 1
	set currently_tipping_ ""
	
	# Bindings for the TipBindings tag.
	bind TipBindings <Enter> {
		%W enter
	}
	bind TipBindings <Leave> {
		TipWidget cancel
	}
	bind TipBindings <Button-1> {
		TipWidget cancel
	}
	bind TipBindings <space> {
		TipWidget cancel
	}
}


# -- proc -------------------------------------------------------------------
# enable
#
# With boolean true, enable all popup tips.  False, disable all of them.
# ---------------------------------------------------------------------------

TipWidget proc enable { {bool {}} } {
	$self instvar currently_tipping_ enabled_

	if { $bool=={} } { return $enabled_ }
	if {! $bool && $currently_tipping_ != ""} {
		$self cancel
	}
	set enabled_ $bool
}


# -- instproc --------------------------------------------------------------
# init
#
# the widget constructor (note that the widget and class arguments are in
# reverse order from "TipWidget create")
# --------------------------------------------------------------------------
TipWidget instproc init { widget cl args } {
	$self set class_ $cl
	eval [list $self] next [list $widget] $args
}


# -- instproc --------------------------------------------------------------
# destroy
#
# the widget destructor: remove the tip, if any
# --------------------------------------------------------------------------
TipWidget instproc destroy { } {
	$self remove_tip
	$self next
}


# -- instproc --------------------------------------------------------------
# create_root_widget
#
# create the widget itself
# --------------------------------------------------------------------------
TipWidget instproc create_root_widget { widget } {
	$self instvar class_
	$class_ $widget
}


# -- instproc ---------------------------------------------------------------
# enter
#
# This is an internal method that is invoked when the <Enter> event occurs
# ---------------------------------------------------------------------------
TipWidget instproc enter { } {
	TipWidget instvar enabled_ currently_tipping_ after_id_
	if {! $enabled_ } continue
	
	set tip [$self make_tip_window]
	if { $tip!="" } {
		wm withdraw $tip
		set delay [option get $tip windowdelay TipWidget]
		if { $delay=="" } { set delay 500 }
		set after_id_ [after $delay \
				"catch {wm deiconify $tip}"]
	}
}


# -- proc -------------------------------------------------------------------
# cancel
#
# Cancel any currently displayed tip by unmapping the window (if it's mapped)
# and by canceling the timer to display a window.
# ---------------------------------------------------------------------------
TipWidget proc cancel { } {
	$self instvar enabled_ currently_tipping_ tip_ after_id_

	if {! $enabled_ } return

	if {$currently_tipping_ != ""} {
		catch {destroy $tip_}
		set currently_tipping_ ""
		set tip_ ""
	}
	catch {after cancel $after_id_}
}


# -- instproc ---------------------------------------------------------------
# make_tip_window
#
# Make and map the tip window for the given widget.
# ---------------------------------------------------------------------------
TipWidget instproc make_tip_window { } {
	TipWidget instvar currently_tipping_ tip_

	set msg [$self cget -tip]
	if { $msg=="" } return ""

	# cancel any previous tip
	TipWidget cancel

	# Note that we're mapping the tip window.
	set currently_tipping_ [$self info path]	

	# create the tip window and prevent the window manager from slapping a
	# border around it
	set count 0
	while { [winfo exists .tip$count] } { incr count }
	set tip_ [toplevel .tip$count -class TipWidget -bd 0 -relief flat]
	wm overrideredirect $tip_ yes

	# Figure out the X and Y coord of the tip window.
	# It should be below the widget it's tipping,
	# and slightly to the right of its left edge, too.
	set yoffset [option get $tip_ yoffset TipWidget]
	if { $yoffset=="" } { set yoffset 2 }
	set xoffset [option get $tip_ xoffset TipWidget]
	if { $xoffset=="" } { set xoffset 20 }
	
	set y [expr [winfo rooty $currently_tipping_] \
			+ [winfo height $currently_tipping_] + $yoffset]
	set x [expr [winfo rootx $currently_tipping_] + $xoffset]
	
	# set the position to the tip window
	wm geometry $tip_ +$x+$y 

	# Insert the message and we're done.
	pack [message $tip_.message -text $msg]
}


TipWidget instproc config_tip { option tip } {
	$self instvar tip_

	if { $tip=={} } {
		$self remove_tip
	} else {
		$self add_tip $tip
	}
}


TipWidget instproc cget_tip { option } {
	$self instvar tip_
	if [info exists tip_] {
		return $tip_
	} else {
		return ""
	}
}


# -- instproc ---------------------------------------------------------------
# add_tip
#
# Add a tip message to the given widget.  When the pointer enters the
# widget, it'll wait a bit, then if it's still there, it'll map the
# tip window.  As soon as the pointer leaves, then it unmaps it.
# ---------------------------------------------------------------------------

TipWidget instproc add_tip { tip } {
	$self instvar tip_
	
	# Save the message text for later use.
	set tip_ $tip
	
	# Add the tip bindings to the window ... now, it might already have
	# them if we got called to change the text, so check first.
	set window [$self info path]
	set tags [bindtags $window]
	if {[lsearch -exact $tags TipBindings] == -1} {
		# They're not in there, add 'em.
		bindtags $window [concat TipBindings [bindtags $window]]
	}
}


# -- instproc ---------------------------------------------------------------
# remove_tip
#
# Remove the popup tip for this window.
# ---------------------------------------------------------------------------
TipWidget instproc remove_tip { } {
	$self instvar tip_
	TipWidget instvar currently_tipping_

	# Do we have saved text for this window?  If so, then we have to get
	# rid of it and the window's class binding, too.
	if [info exists tip_] {
		# Okay, nuke the text.
		unset tip_
		
		# Now, get the tags and see if the TipBindings is among them.
		set window [$self info path]
		set tags [bindtags $window]
		if {[set index [lsearch -exact $tags TipBindings]] >= 0} {
			# Yep, it's there.  Set the new bindings without
			# TipBindings present.
			bindtags $window [lreplace $tags $index $index]
		}
		
		# Also, if the tip window is visible for this widget, nuke it.
		if {$currently_tipping_ == [$self info path]} {
			TipWidget cancel
		}
	}
}



# initialize the TipWidget class
TipWidget init_

